/*
  This program is free software published under the
  terms of the GNU Lesser General Public License.
  http://www.gnu.org/copyleft/lesser.html    
  This program is distributed in the hope that it will
  be useful - WITHOUT ANY WARRANTY; without even the
  implied warranty of MERCHANTABILITY or FITNESS FOR A
  PARTICULAR PURPOSE.
*/ 


/*  xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
     source code: ES6 (ECMAScript 6, ECMAScript 2015, JavaScript 6)
     author: kekec 		E-mail: behaviour.brain.research@gmail.com
		xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */

 
//                          lst version: 1.0


//  xxxxxxxxxxxxxxx    global variable declarations  xxxxxxxxxxxxxxxx

 
// These variables below can be directly modified by the user on the options form: 

var stype = '2';  // Stimulus type: '1' = Shapes, '2' = Colors, '3' = Letters, '4' = Numbers, '5' = Numbers-sudoku

var row_min = 5;  // Minimum number of rows and columns

var row_max = 7; // Maximum number of rows and columns

var iter = 1;   // Number of iterations for each row value

var timer_enabled = false;   // if true => timer is enabled. if false timer is disabled. Exactly: the timer is always on

var timer = 600;     // Task time in seconds if clock = 1000ms (default). 

var font_family = 'Arial, Helvetica, sans-serif';  // default canvas ctx.fontFamily 

var is_back = true;    // if (is_back==true) => user can go back to previous task/item.

var auto_f = false;    // Auto-forward. Automatically move to the next task (true || false) 

var prg_type = '0';   //  type of information in Progress Bar ('0' => Items, '1' => Time; '2' => Nothing) 

var is_shuffle = false;   // Randomly shuffle rows (only when Demo mode = Off)

var demo = false;       // if (demo == true) => Show task solutions 

var back_clr = '#fff';  // Background color during task

var pen_clr = '#000';  //  Color for geometric shapes (numbers, letters)

var is_border = true;   // Draw border around numbers

var store = false;      // false => this web application will not store results within client web browser.; true => This web application will store the results inside client web browser. 

// Other global variables:

const clock = 1000;  // basic timer resolution in ms

var timer_id = null;   // timer_id is the identifier of the repeated action you want to cancel. This ID is returned from setInterval().

var counter = 0;      // timer counter. if (counter >= time) Timer will call theEnd

var item_cnt;       // number of items in the task. range [0..item_cnt-1]. see setPar()

var item;   				// current item in task

var is_task = false;    // task in progress or not for pageResize() event. It is true for all events between end of startTask() and theEnd() events

var dirty = false;  	// if  (dirty==true) => user changed any options in form 

var is_storage = false;		// local storage is enabled on user browser or not
  
//  global arrays declarations..
 
var stim = [];   // stimulus array (user answer array). [items] x [rows] x [cols].  see stimGen()   


/*  xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
    xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
		xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */


// application init [create/start Timer, check storage, show start panel..]
function init() 
{ 
  if (timer_id)  {   
    window.clearInterval(timer_id);  
  }  
  timer_id= setInterval(function () {appTimer()},clock);  
	
	is_storage= isStorage() && storageAvailable('localStorage');
  if (is_storage)  {
    document.getElementById('store_group').style.display= 'block';
		document.getElementById('no_storage').style.display= 'none';
  }	
	
	loadFont();
	
	setIns();
	
	showStart();  
}

 
// application timer handler function
function appTimer()  
{    
	if (timer_enabled && is_task)   // else: nop, idle, do nothing 
	{
		counter++;  
		if (counter >= timer)  {     // task time is limited
			theEnd(); 
		}	
		else if (prg_type == '1')   {   // '1' => Time. display time on Progress Bar  
			let rest= timer-counter;   // this is remaining time 
			let percent= Math.round(100*rest/timer);    
			document.getElementById('progress').setAttribute('style', 'width:'+percent+'%');    // !important
			document.getElementById('nav_items').textContent= stringTime(rest);	
		}
	}  
}	


// before using web storage, check browser support for localStorage and sessionStorage
function isStorage()  
{
  if (typeof(Storage) !== 'undefined') {
     return true;                 
  } 
  else {
    return false;    
  }  
}


// necessarily because of IE11 (probably redundant)
function storageAvailable(type) 
{
  try  {
    return (window[type]) ? true: false;   
  }
  catch(e) {
    return false;
  }
}


// return integer into string hh:mm:ss time format
// input par: total == integer (time in seconds)
function stringTime(total) 
{
  let hh = Math.floor(total/(60*60));         
  let div_min = total % (60 * 60);
  let mm = Math.floor(div_min / 60);       
  let div_sec = div_min % 60;
  let ss = Math.ceil(div_sec);    
  // These lines ensure you always have two-digits. This formats your string to HH:MM:SS
  if (hh < 10)  {hh = "0"+hh;}     
  if (mm < 10) {mm = "0"+mm;}
  if (ss < 10) {ss = "0"+ss;} 
  return hh + ":" + mm + ":" + ss;
} 


// fill font-family names from array into select html element
function loadFont()  
{
	// used font-family array: The name of a font family. For example, "Times" and "Helvetica" are font families. Font family names containing whitespace should be quoted
	// https://www.w3schools.com/cssref/css_websafe_fonts.asp	
	const arr = ['"Times New Roman", Times, serif', 'Georgia, serif', '"Palatino Linotype", "Book Antiqua", Palatino, serif', 'Arial, Helvetica, sans-serif', '"Arial Black", Gadget, sans-serif', '"Comic Sans MS", cursive, sans-serif', 'Impact, Charcoal, sans-serif', '"Lucida Sans Unicode", "Lucida Grande", sans-serif', 'Tahoma, Geneva, sans-serif', '"Trebuchet MS", Helvetica, sans-serif', 'Verdana, Geneva, sans-serif', '"Courier New", Courier, monospace', '"Lucida Console", Monaco, monospace' ];
	let sel= document.getElementById('font_family'), len = arr.length, opt; 
	sel.innerHtML= ''; 
	for (let i= 0; i!= len; i++) 
	{  
		opt = document.createElement('option');
		opt.value= arr[i];  // option value selected font-family name (string)            
		opt.text= arr[i];   // option textContent => font-family name (string)              		                       								
		sel.add(opt);  // sel.add(opt, null);                 			
	} 
	sel.selectedIndex= 3; // 'Arial, Helvetica, sans-serif'
	document.getElementById('font_family_msg').style.fontFamily= sel.value; // = arr[3];
}	


// raised when page is resized. see html: <body onresize="pageResize()"
function pageResize()  
{ 
	if ( is_task )  { 
		document.getElementById('can_m').style.display= 'none';  // tmp until mouseMove()
		setCanvas();		
		drawStim();		
	}
}	


// show start panel 
function showStart()  
{     
  document.getElementById('end').style.display= 'none';
  document.getElementById('result_wrapper').style.display= 'none';  
  document.getElementById('store_id').style.display= 'none';
	document.getElementById('table_grp').style.display= 'none';
	document.getElementById('start').style.display= 'block';	
}


// show form with application options
// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
function showOption()   
{
  document.getElementById('hide_option').style.display= 'block';
  document.getElementById('options').style.display= 'block';
  document.getElementById('start').classList.remove('center-xy');  // ex 'child' class    
  document.getElementById('start_div').style.display= 'none';    
  document.getElementById('show_option').style.display= 'none';  
  document.getElementById('hide_option').style= 'text-align:center';  // bug? auto-remove style when element.style.display="none"
  document.getElementById('instructions').style.display= 'none';     
  document.getElementById('store_oper').style.display= 'none';
	if (is_storage && localStorage.getItem('lst-v1.0')) 
  {
    let cnt= JSON.parse(localStorage.getItem('lst-v1.0')).length;                                                        
    document.getElementById('store_cnt').textContent=  'You have a total of ' + cnt  + ' results stored in your web browser.';
    if (cnt > 0) {
			document.getElementById('store_oper').style.display= 'table';  // xxx
		}
	}
  dirty= false;
  //document.getElementById('is_valid').textContent='';  // tmp reset validator message
}


// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
// save to window global vars all user Options/Settings
function hideOption()  
{
  setVarPar();   
  document.getElementById('options').style.display = 'none'; 
  document.getElementById('start').classList.add('center-xy');   // ex 'child' class 
  document.getElementById('show_option').style.display = 'block';   
  document.getElementById('start_div').style.display = 'block';          
  document.getElementById('hide_option').style.display = 'none';   
  //document.getElementById('start_div').style= 'text-align:center';  // bug?  (firefox auto remove class "center" when set style.display.none;)
  document.getElementById('show_option').style= 'text-align:center';  
  document.getElementById('instructions').style.display= 'block';
	refMsg(); 
  setIns();	
}


// set instructions text depending on user selected task options
function setIns()  
{
  document.getElementById('instructions').style.display= 'block';		

	document.getElementById('ins_back').style.display= is_back ? 'none' : 'block';
	document.getElementById('ins_demo').style.display= demo ? 'block' : 'none';
	
	if ( document.getElementById('timer_enabled').checked )
  {   
    document.getElementById('ins_timer').setAttribute('style', 'display:block'); 		
		document.getElementById('ins_timer').textContent= 'The duration of this task is limited.';  // The duration of this task is always limited.
		if (prg_type == '0')  {
			document.getElementById('ins_timer').textContent+= ' The Progress bar will show how many items are left until the end of the task.';
		}
		else if (prg_type == '1')  {
			document.getElementById('ins_timer').textContent+= ' The Progress bar will display the remaining time until the end of the task.';  
		}
		else  {
			document.getElementById('ins_timer').textContent+= ' But information about the remaining time will not be visible.';  
		}  
  }
	else  {
		document.getElementById('ins_timer').setAttribute('style', 'display:none');
	}	
	
	let e= document.getElementById('stype'), ind= e.selectedIndex, v = e.options[ind].textContent;
	if (ind == 4) {
	  v = e.options[3].textContent;
		document.getElementById('ins_sudoku').style.display= 'block';
	}
	else  {
		document.getElementById('ins_sudoku').style.display= 'none';
	}
	document.getElementById('ins_stype').textContent= v;
	
}	


// refresh Options (update *_msg label) => give user short option HELP
function refOpt(e)  
{   
  dirty = true;       
  let my_id = e.id, msg = document.getElementById(my_id + '_msg'), s = 'Values within this range are allowed: ', s2= ' with step= ';               
	switch (my_id)
	{
		
		case 'stype':
		  let ind = e.selectedIndex, min_r= document.getElementById('row_min'), max_r= document.getElementById('row_max');
			document.getElementById('rows_div').style.display =  ind == 4 ? 'none': 'block';
			msg.textContent= e.options[ind].getAttribute('data-help'); 
			switch (ind) {
				case 0: 
					min_r.max= 37;  
					max_r.max= 37;  // shapes. see drawShape() + genItem()
					break;
				case 1: 
					min_r.max= 16;  
					max_r.max= 16;  // colors. // clr.length= 16. see: stimGen() + getColor()
					break;
				case 2: 
					min_r.max= 26;
					max_r.max= 26;  // letters = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ';
					break;							
				case 3: 
					min_r.max= 50;
					max_r.max= 50;  // numbers
					break;	

				/*case 4: 
					min_r.max= 9;
					max_r.max= 9;  // sudoku
					break;*/	
			}			
			checkItem();
			break;
	
		case 'row_min':
		case 'row_max':
		case 'iter':
			if (e.validity.valid)  {
				msg.textContent= ''; 				
			}
			else  {
				msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
			} 
			checkItem();
			break; 
   	  
    case 'timer_enabled':   
      if (e.checked)  {
        msg.textContent= 'Yes.';
        document.getElementById('p_timer').style.display= 'block';
      }
      else  {  
        msg.textContent= 'No. Timer is disabled.';
        document.getElementById('p_timer').style.display= 'none'; 
      }       
      break;
		
    case 'timer':
			if (e.validity.valid)  {
				msg.textContent= stringTime(e.value); 
				msg.removeAttribute('class');
			}
			else  {
				msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');
			} 
			break; 

		/* case 'font_family':
		  break; */
		
		case 'is_back':		
			msg.textContent= e.checked ? 'Yes.' : 'No.';      
			break; 
			
		case 'auto_f':		
			msg.textContent= e.checked ? 'When the item is successfully completed, feedback is received (Alert dialog box) and the next item is automatically displayed.' : 'No.';      
			break; 
			
		case 'prg_type':                                                                                                                                            
      if (e.value == '0')  {
        msg.textContent= 'It will be displayed how many items are left to the end.'; // The number of the current item will be displayed, as well as the total number of items.
      }
      else if (e.value == '1')  {
        msg.textContent= 'The remaining time until the end of the task will be displayed. Only when timer is enabled.';
      }
      else  {
        msg.textContent= ''; //'Nothing';
      }
      break;    

    case 'is_shuffle':		
			msg.textContent= e.checked ? 'Yes. The elements of the matrix will initially be randomly mixed. There is a low probability that the elements of the matrix are placed in the correct position, especially if the number of rows of the matrix is small.' : 'No. Equal elements will be initially placed in each column of the matrix.';      
			break; 

		case 'demo': 
			msg.textContent= e.checked ? 'On. The solution of each task will be displayed.' : 'Off';  
			document.getElementById('p_is_shuffle').style.display= e.checked ? 'none' : 'block';
			break; 
		
		case 'back_clr': 
		case 'pen_clr':  		  
			msg.textContent= ( validColor(e.value) ) ? '' : 'The selected color is invalid.'; 
			break; 	

    case 'is_border':		
			msg.textContent= e.checked ? 'Yes.' : 'No.';      
			break; 

		case 'store':                                                                                                                                            
			msg.textContent= (e.value === '1') ? 'This web application will store the results inside your web browser (no expiration date). Useful for group research' : 'This web application will not store results within your web browser. Data is lost when the browser tab is closed';
			break;
 
	}   // switch
	
  refMsg();   
}


// check if item count is valid or not
// update item count on options form and check row_min <= row_max and update item_total label
function checkItem()  
{
  let row_min = document.getElementById('row_min');
  let row_max = document.getElementById('row_max');
  let iter = document.getElementById('iter');
  let items = document.getElementById('item_total');
  let row_min_msg = document.getElementById('row_min_msg');
  let row_max_msg = document.getElementById('row_max_msg');    
  if (row_min.validity.valid && row_max.validity.valid && iter.validity.valid)   
  {                      
    if  ( parseInt(row_min.value,10) > parseInt(row_max.value,10) )  
    {
			let tmp= 'Minimum number of rows must not be higher than the Maximum number of rows. The starting level must not be higher than the final level.';
      row_min_msg.textContent= tmp;  
      row_max_msg.textContent= tmp;  
      items.textContent= 'Total number of items = Invalid value'; 
      items.setAttribute('class', 'error');
    } 
    else   
    {                                                                                               
      row_min_msg.textContent= '';  
      row_max_msg.textContent= '';
      let cnt= 0, level= parseInt(row_min.value,10);  // update cnt => total item count in task  
      while (level <= parseInt(row_max.value,10))  {
        cnt+= parseInt(iter.value,10);    
        level++;   
      } 
      items.textContent=  'Total number of items = ' + cnt ;      
      items.removeAttribute('class');
    }
  } 
  else  
  {
		let e= document.getElementById('stype'), my_type= e.options[e.selectedIndex].textContent, s = ' values within this range are allowed: ', s2= ' with step= ';     
    row_min_msg.textContent= 'For selected Stimulus type: ' + my_type + s + row_min.min + ".." + row_min.max + s2 + row_min.step;
		row_max_msg.textContent= 'For selected Stimulus type: ' + my_type + s + row_max.min + ".." + row_max.max + s2 + row_max.step;
		items.textContent= 'Total number of items = Invalid value'; 
    items.setAttribute('class', 'error');
  }    
}


// refresh main validator message for user defined OPTIONS 
function refMsg()   
{ 	
	if ( optionsValid() )  {
    document.getElementById('is_valid').textContent=  dirty === true ? 'All your settings have been successfully updated' : '';  
    document.getElementById('is_valid').setAttribute('style', 'text-align:center;color:#000000'); 
  }
  else  {
    document.getElementById('is_valid').textContent='Some of your options are incorrectly set. They were not updated.';   // These settings are marked in red. 
    document.getElementById('is_valid').setAttribute('style', 'text-align:center;color:#ff0000;font-weight:bold'); 
  } 	
}


// check for validity most of options 
function optionsValid()  
{ 
  let res = document.getElementById('row_min').validity.valid 
				 && document.getElementById('row_max').validity.valid
	       && document.getElementById('iter').validity.valid 			 
				 && document.getElementById('timer').validity.valid  	 
				 && validColor(document.getElementById('back_clr').value)
				 && validColor(document.getElementById('pen_clr').value); 
	
	if ( parseInt(document.getElementById('row_min').value,10) > parseInt(document.getElementById('row_max').value, 10) )  {  
		res = false;
	}		
  return res;              
}


// return random integer in range [min..max]
function getRndInt(min, max) {
	return Math.floor(Math.random() * (max - min + 1) ) + min;
}


// start task (research) 
function startTask(e)   
{      			
  initLayout(); 	
	
  setPar();
 
	stimGen();
	
	is_task= true;   // if ( timer_enabled ) finally "activate" timer 
	
	drawStim();    // draw first item (stimulus)
	
}


// init DOM/html elements (positions,visibility,..)
function initLayout()
{	
	document.getElementById('result_wrapper').style.display= 'none';
  document.getElementById('end').style.display= 'none'; 
  document.getElementById('start').style.display= 'none';              
	document.getElementById('task_section').style.display= 'block'; 	  	
	
 	document.getElementById('progress').style.display= 'block';   //document.getElementById('progress').setAttribute('style', 'width:100%');	 
	if ( document.getElementById('prg_type').value == '2' || ( document.getElementById('prg_type').value == '1' && timer_enabled == false ) ) {
		document.getElementById('progress_holder').style.display= 'none'; 
	}	
	else  
	{
		document.getElementById('progress_holder').style.display= 'block';     
		//document.getElementById('progress').setAttribute('style', 'width:100%');   		
		if (document.getElementById('prg_type').value == '0') {
			document.getElementById('progress').className='progress-bar progress-bar-info progress-bar-striped';     //  items / remaining items
		}
		else if (document.getElementById('prg_type').value == '1')  {
			document.getElementById('progress').className= 'progress-bar progress-bar-danger progress-bar-striped';  // remaining time   
		} 	
		document.getElementById('progress').setAttribute('style', 'width:0');  // reset progress bar width
	}					
	document.getElementById('nav_items').textContent='';  // always reset  

  setCanvas();
	
}  


// set html canvas element absolute positions
function setCanvas()
{
	// canvas must be after set footer content => progress_holder visible or not
	let c= document.getElementById('can_s'), ctx= c.getContext('2d');
	c.style.display = 'block';   
	c.width = innerWidth;  
	c.style.width = c.width + 'px';	     
	c.height = innerHeight - document.getElementById('footer').offsetHeight; // document.getElementById('footer').clientHeight == document.getElementById('footer').offsetHeight  
	c.style.height = c.height + 'px';  
	
	ctx.textAlign = 'center';    // ctx.textAlign = "left" || "right" || "center" || "start" || "end";   Default value: 	start
	ctx.textBaseline = 'middle';  // const baselines = ['top', 'hanging', 'middle', 'alphabetic', 'ideographic', 'bottom'];  Default value: alphabetic
 
	ctx.strokeStyle= pen_clr;  // for draw border in drawStim() => equal to fillStyle.
  ctx.lineWidth= 1; 	
	
}


// set internal application parameters (global vars)
function setPar()  
{  
	if ( stype == '5' )  {  // fixed
		row_min= 9;
		row_max= 9;
		document.getElementById('row_min').value= 9;
		document.getElementById('row_max').value= 9;
	}
	if ( row_min > row_max ) {  // this should be impossible but because of danger of indefinite loop in stimGen()
    let v = Math.min(row_min, row_max);
		row_min= v;
		row_max= v;
		document.getElementById('row_min').value= v;
		document.getElementById('row_max').value= v;
	}		
	
	item = 0; 
	counter= 0;       // reset timer counter		
	stim = [];	      // reset stimulus array 
	
}


// generate all stimuli for each item in range 0..item_cnt-1	
function stimGen()	
{		
	let row= Math.min(row_min, row_max), end= Math.max(row_min, row_max);  
	while ( row <= end )  
	{
		for (let j= 0; j< iter; j++) {    
			genItem(row);
		}
		row++;
	} 	
	
	item_cnt= stim.length;   // set item_cnt => total items count in task 
}


/* generate new stim[item] content   
   par rows = number of rows = number of columns */
function genItem(rows)  
{
	let col= rows, shapes= {}, clr = {}, result = [];
	for (let i= 1; i<= rows; i++)  
	{  				
		let r, row= []; 		
		switch (stype)
		{
			case '1':  // Shapes
				do	{
					r = getRndInt(0,36);
				} while ( shapes[r] == 1 );
				shapes[r]= 1;			
				for (let i= 0; i < col; i++)  {
					row.push(r);
				}
				break;
				
			case '2':  // Colors (default)
			  do  {
					r = getColor();
				} while (clr[r] == 1);
				clr[r]= 1;
				for (let i= 0; i < col; i++)  {
					row.push(r);
				}
				break;
			
			case '3':   // Letters
				let letters = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ';
				let txt = letters.substr(0, col); 
				for (let i= 0; i < col; i++)  {
					row.push(txt[i]);
				}
				break;
		
			case '4':  // Numbers
			case '5':  // Numbers-sudoku
				for (let i= 1; i <= col; i++)  {
					row.push(i);
				}
				break;
		
		}

		result.push(row);			
	} 

	if ( stype == '1' || stype == '2' )  {  // values within each columns must be const
		transposeArray(result);
	}
		
	let size = result.length;
	if ( demo )  
	{  // set correct solution
		if ( stype == '5' )  {
			let offset = [ 0, 3, 6, 1, 4, 7, 2, 5, 8 ];  // shift (offset) "right" for each row: 0..8
			for (let i= 1; i < size; i++)  {   
				result[i] = shift(result[i], offset[i]);
			}
		}
		else  
		{
			for (let i= 0; i < size; i++)  {   
				result[i] = shift(result[i], i); 
			}
		}
	}		
  else if ( is_shuffle ) 
	{ // shuffle rows (only if demo mode = false) 
		for (let i= 0; i < size; i++)  {   
			result[i] = shuffle(result[i]); 
		}
	}	

	stim.push(result);   // xxxx  

}


// sets one of mostly basic html predefined colors
// this is mostly default set of 16 html Basic colors 
function getColor() 
{
  let clr = [    
    '#000000',     /* "black" */   
    '#FF0000',     /* "red"   */
    '#008000',     /* "green" */
    '#0000FF',     /* "blue" */   
    '#FFFF00',     /* "yellow" */     
    '#00FF00',     /* "lime" */     
    '#00FFFF',     /* "cyan" / "aqua" */     
    '#FF00FF',     /* magenta / fuchsia  */  
    '#C0C0C0',     /* "silver" */      
    '#800000',     /* maroon */     
    '#808000',     /* "olive" */  
    '#800080',     /* "purple" */    
    '#008080',     /* teal */      
    '#000080',     /* navy */    
    '#FFC0CB',     /* "pink" */           
    '#FFA500'      /* "orange" */ 
    ];      
  //let rnd= getRndInt(0,15);  // array clr limits
  let rnd=  Math.floor(Math.random()*16);   // 0..15
  return clr[rnd];
}


/* set correct solution
   arr = input 1-dim array
	 start = start point for "right" shift/offset */
function shift(arr, start)
{
  let ind= start, up= arr.length-1, res=[];
	do  {
		res.push(arr[ind]);
		ind++; 
		if ( ind > up ) {
			ind = 0;
		}	
	} while (ind != start);
	return res;
}	


// IN PLACE change rows and columns valuse in 2-dim array (matrix)
function transposeArray(matrix) 
{
	let size = matrix.length, tmp;
  for (let i = 0; i < size; i++)  {
    for (let j = 0; j < i; j++)  {
      tmp = matrix[i][j];
      matrix[i][j] = matrix[j][i];
      matrix[j][i] = tmp;
    }
  }
}


/* 	change rows and columns valuse in 2-dim array (matrix)
		return new array. don't change original array (matrix) 
function transpose(matrix) 
{
	let rows= matrix.length, cols= matrix[0].length, res= [];
	for (let j = 0; j < cols; j++){
		res.push([]);
	}	
	for (let i = 0; i < rows; i++)  
	{
    for (let j = 0; j < cols; j++)  {
      res[j][i]= matrix[i][j];
    }
  }
	return res;
}   */

  
// unbiased shuffle algorithm: Fisher-Yates (aka Knuth) Shuffle.
function shuffle(arr) 
{
  let cur_index = arr.length, tmp, rnd_index;
  // While there remain elements to shuffle...
  while (0 !== cur_index) {    
    rnd_index = Math.floor(Math.random() * cur_index);     // Pick a remaining element...
    cur_index = cur_index - 1;   
    tmp = arr[cur_index];          // And swap it with the current element.
    arr[cur_index] = arr[rnd_index];
    arr[rnd_index] = tmp;
  }
  return arr;
}


// draw (repaint) current stimulus           
function drawStim() 
{    		 
	let can = document.getElementById('can_s'), ctx = can.getContext('2d'), width = can.width, height = can.height, k, sud_clr;					
	ctx.fillStyle= back_clr;
	ctx.fillRect(0, 0, width, height);    // ctx.clearRect(0, 0, w, h);
  ctx.fillStyle= pen_clr;    	 //ctx.lineWidth= xxx; 

  let row_cnt= stim[item].length;
	let col_cnt= stim[item][0].length;
	width  /=  col_cnt;  // cell width
  height /=  row_cnt;  // cell height
	let r = Math.floor( Math.min(width,height)/2 );  
	ctx.font = r.toString() + 'px ' + font_family;   	// update font		

	if ( stype == '5' )  {
		sud_clr = ['#FF0000','#00FF00', '#0000FF', '#FFFF00','#FFC0CB', '#808000', '#800080', '#00FFFF', '#FFA500' ]; // array for sudoku background-colors
		k= 0;  // sudoku background-color array counter
	}
	for (let row = 0; row < row_cnt; row++)
	{	   	
		for (let col= 0; col < col_cnt; col++)  
		{                						
			let left= col * width;   // cell left bound
			let top = row * height;  // cell top bound      
			
			switch  ( stype  )  
			{  
				case '1':  // Shapes
					drawShape(ctx, Math.round(left + width/2), Math.round(top + height/2), Math.floor(r-2), stim[item][row][col], pen_clr); // function drawShape(ctx, x, y, r, shape, col)			
					break; 
			
			  case '2':  // Colors
					ctx.fillStyle = stim[item][row][col]; 
					ctx.fillRect(left, top, width, height);   // void ctx.fillRect(x, y, width, height);   
					break;
			
				case '3':  // Letters
				case '4':  // Numbers
					ctx.fillText(stim[item][row][col], Math.round(left+width/2), Math.round(top+height/2));				
					break;
					
				case '5':  // Numbers-sudoku					
					if ( ( row == 0 || row == 3 || row == 6 ) &&  ( col == 0 || col == 3 || col == 6 ) )  {					  
						ctx.fillStyle = sud_clr[k];
						ctx.fillRect(left, top, 3*width, 3*height);   // void ctx.fillRect(x, y, width, height); 
						k++;		
					}					
					ctx.fillStyle= '#000'; 
					ctx.fillText(stim[item][row][col], Math.round(left+width/2), Math.round(top+height/2));				
					break;						
			}
			
      if ( is_border )	{
				ctx.beginPath();  
				ctx.rect(left, top, width, height);     
				ctx.stroke(); 	
			}			
		}
	} 

	updateFooter();	

}


/* shape function dispatcher (handler)
   ctx = canvas context
   x = x coord rect center,
   y = y coord rect center,
   r = shape radius,   
   shape = shape type => this argument is main for call different shapes.js functions  
   col = color (fillSyle and/or strokeStyle)
   shapes.js file must be included in DOM  (in html file)
	 draws all P-type stimuli     */
function drawShape(ctx, x, y, r, shape, col)
{ 
  ctx.fillStyle= col;    // fill color
  ctx.strokeStyle= col;  // line color   
  switch (shape)  
  {
    case 0:   // full circle
    case 1:   // 1/2 circle RIGHT oriented
    case 2:   // 1/2 circle DOWN oriented   
    case 3:   // 1/2 circle LEFT oriented
    case 4:   // 1/2 circle UP oriented 
    case 5:   // 3/4 circle with UP-RIGHT hole
    case 6:   // 3/4 circle with DOWN-RIGHT hole
    case 7:   // 3/4 circle with DOWN-LEFT hole 
    case 8:   // 3/4 circle with UP-LEFT hole                          
     if (shape == 0)  {                        
       circle(ctx, x, y, r, 0, 2*Math.PI, false, col);  // full circle 
     }        
     else if (shape == 1)  {
       circle(ctx, x, y, r, 1.5*Math.PI, 0.5*Math.PI, false, col); // 1/2 circle RIGHT oriented      
     }
     else if (shape == 2)  {
       circle(ctx, x, y, r, 0, Math.PI, false, col); // 1/2 circle DOWN oriented     
     } 
     else if (shape == 3)  {
       circle(ctx, x, y, r, 0.5*Math.PI, 1.5*Math.PI, false, col); // 1/2 circle LEFT oriented     
     } 
     else if (shape == 4)  {
       circle(ctx, x, y, r, Math.PI, 0, false, col);   // 1/2 circle UP oriented     
     }        
     else if (shape == 5)  {
       circle(ctx, x, y, r, 0, 1.5*Math.PI, false, col);   // 3/4 circle with UP-RIGHT hole     
     } 
     else if (shape == 6)  { 
       circle(ctx, x, y, r, 0, 0.5*Math.PI, true, col);   // 3/4 circle with DOWN-RIGHT hole     
     } 
     else if (shape == 7)  { 
       circle(ctx, x, y, r, 0.5*Math.PI, Math.PI, true, col);   // 3/4 circle with DOWN-LEFT hole     
     } 
     else /*if (shape == 8)*/  { 
       circle(ctx, x, y, r, 1.5*Math.PI, Math.PI, false, col);   // 3/4 circle with UP-LEFT hole     
     }        
     break;
                
    case 9:   // polygon 3 equal sides
    case 10:  // polygon 4 equal sides   
     (shape == 9) ? polygon(ctx, x, y, r, 3, col):  polygon(ctx, x, y, r, 4, col);       
     break;    
                  
    case 11:  //  xxxxxx
     square(ctx, x, y, r-2, col);  // only for square: window.r must be > 2 for shape selection and demo background fill style  
     break;  //   xxxxxx
             
    case 12:    // longer side (base) down
    case 13:    // longer side (base) up
     (shape == 12) ? trapezoid(ctx, x, y, r, col, true) : trapezoid(ctx, x, y, r, col, false) ;  // base down or up   
     break;
              
    case 14: // rectangle - horizontal line is longer then vertical
    case 15: // rectangle - vertical line is longer then horizontal 
     (shape == 14) ? rectangle(ctx, x, y, r, col, true): rectangle(ctx, x, y, r, col, false);              
     break;
                   
    case 16:
    case 17:
     let w, h;
		 if (shape == 16) {   // vertical ellipse
       w= r, h= 2*r; // change this 2 param for PI/2 rotation
     }
     else  {   // horizontal ellipse
			 w= 2*r, h= r; // change this 2 param for PI/2 rotation
     }
     ellipse(ctx, x - w / 2.0, y - h / 2.0, w, h, col);       
     break; 
                
    case 18: // cross => classic vertical cross
    case 19: // cross => diagonal cross (x) 
     (shape == 18) ? cross(ctx, x, y, r, col, true): cross(ctx, x, y, r, col, false);       
     break;   
             
    case 20:  //(base == true) => draw down-right rotated paralelogram; 
    case 21: // (base== false) => draw left-down rotated paralelogram
     (shape == 20) ? paralelogram(ctx, x, y, r, col, true): paralelogram(ctx, x, y, r, col, false);    
     break;
                               
    case 22:   // star 4-points
    case 23:   // star 5-points
    case 24:   // star 6-points     
     star(ctx, x, y, shape-18, r/2, r, col);       
     break;
                            
    case 25:  //  vertical UP arrow (0)
    case 26:  //  vertical DOWN arrow (1)
    case 27:  //  horizontal LEFT arrow (2)
    case 28:  //  horizontal RIGHT arrow (3)
     arrow(ctx, x, y, r, col, shape-25);       
     break;
            
    case 29:  
    case 30:
     let a= (shape == 29) ? [r, r, 0, 0] : [0, 0, r, r];    // up rect round radius (31) or down round rect radius (32)
     roundRect(ctx, x-r, y-r, 2*r, 2*r, a, col);    
    break;
       
    case 31:     //  mushroom UP oriented     
    case 32:     //  mushroom RIGHT oriented
    case 33:     //  mushroom DOWN oriented
    case 34:     //  mushroom LEFT orinted 
      mushroom(ctx, x, y, r, col, shape-31);
      break;
               
    case 35:  // polygon 5 equal sides
    case 36:  // polygon 6 equal sides      
      (shape == 35) ?  polygon(ctx, x, y, r, 5, col):  polygon(ctx, x, y, r, 6, col);       
      break; 
                      
  }   // switch (shape) of
   
}  		


// raised when user change font-family on html <select> element
function tryFont(e)  {	
	document.getElementById('font_family_msg').style.fontFamily= e.value;
}	


// Changes application settings that can be changed directly by the user within the options form
// These variables below can be directly modified by the user on the options form
function setVarPar()  
{   

  if ( document.getElementById('stype').selectedIndex >= 0 ) {
		stype = document.getElementById('stype').value;
	}		
	
	if (document.getElementById('row_min').validity.valid && document.getElementById('row_max').validity.valid 
    && (parseInt(document.getElementById('row_min').value,10) <= parseInt(document.getElementById('row_max').value, 10) ) )   {  
      row_min= parseInt(document.getElementById('row_min').value, 10); 
      row_max= parseInt(document.getElementById('row_max').value, 10); 
  }	

	if (document.getElementById('iter').validity.valid)  {  
	  iter = Math.round(parseInt(document.getElementById('iter').value, 10));			
  } 	
	
  timer_enabled= (document.getElementById('timer_enabled').checked) ? true : false;  // timer is enabled or disabled	
	
	if (timer_enabled && document.getElementById('timer').validity.valid) {
		timer= Math.round(parseInt(document.getElementById('timer').value, 10));  // task time (s)
  }	
		
	font_family= document.getElementById('font_family').value;	
  
	is_back= document.getElementById('is_back').checked ? true : false; 
	
	auto_f= document.getElementById('auto_f').checked ? true : false; 
	
	prg_type= document.getElementById('prg_type').value;  
	
  is_shuffle= document.getElementById('is_shuffle').checked ? true : false;   
	
  demo= document.getElementById('demo').checked ? true : false;	
	
	if (validColor(document.getElementById('back_clr').value))  {		
		back_clr= document.getElementById('back_clr').value;
  }
	
	if (validColor(document.getElementById('pen_clr').value))  {		
		pen_clr= document.getElementById('pen_clr').value;
  }
	
  is_border = document.getElementById('is_border').checked ? true : false;
	
  store= (is_storage && document.getElementById('store').value === '1') ? true : false;
  
}


// test color for validity (from string)
// https://stackoverflow.com/questions/8027423/how-to-check-if-a-string-is-a-valid-hex-color-representation/8027444
function validColor(clr){
  return /^#[0-9A-F]{6}$/i.test(clr);
}


// goto previous item (stimulus)
function priorItem()  
{      
  if (item > 0)  {
    document.getElementById('can_m').style.display= 'none';  // raster can be changed in next/prior task
		item--;  
    drawStim();     // draw again previous stimulus
  }         
}


// goto next item (stimulus)
function nextItem()  
{    		
	item++;					
	document.getElementById('can_m').style.display= 'none'; // raster can be changed in next/prior task
	if (item < item_cnt)  {         		 
    drawStim();     // draw next stimulus => range 0..item_cnt-1        
  }
  else  {               
    theEnd();  // end of task if item >= item_cnt 
  }
}


// update footer when show each stimulus in task. update buttons and update progress bar
function updateFooter()
{  
	document.getElementById('prior').style.display= item > 0 && is_back ? 'inline': 'none';  // set footer navigation
	let btn= document.getElementById('next');     
	if (item == item_cnt-1)  {       
		btn.style.backgroundImage=  "url('img/end.png')";  // update 'Next item' image on 'right arrow' button 
		btn.title= 'End of task';
	}
	else  {
		btn.style.backgroundImage= "url('img/forward.png')"; 
		btn.title= 'Go forward';
	}						
  if (prg_type == '0' )  {    // update Progress bar => show current item count and remaining item count on Progress Bar                                                                                                                                                       
		let percent = Math.round(100*(item_cnt-(item+1))/item_cnt);      
		document.getElementById('progress').setAttribute('style', 'width:'+percent+'%');    // !important
		document.getElementById('nav_items').textContent=  (item + 1) + ' / ' + (item_cnt - item-1);  
	}
}


/* set absolute positions (left, top, width, height) for can_m object	
   canvas => mousemove event	=> a pointing device is moved over an element (fired continuously as the mouse moves).
	 this event is fired by html canvas element => document.getElementById('can_s') */
function mouseMove(e)
{  
  if ( is_task ) 
	{   	
		let pos = getCellFromXY(e.offsetX, e.offsetY);  // The offsetX (offsetY) property returns the x (y) -coordinate of the mouse pointer, relative to the target element. The clientX (clientY) property returns the horizontal (vertical) coordinate (according to the client area) of the mouse pointer when a mouse event was triggered.
		let row = pos[0], col = pos[1];  // stimulus row and col index   
		
		let can_s= document.getElementById('can_s'), raster= stim[item].length;
		let dx =  can_s.width / raster, dy = can_s.height / raster;  // dx= cell (can_m) width;  dy = cell (can_m) height
		
		// set absolute positions (left, top, width, height) for m_canvas object			
		let can_m = document.getElementById('can_m');
		can_m.style.left= col * dx  + 'px';  // top and left properties require a unit of measure, not just a number
		can_m.style.top=  row * dy + 'px';   
		can_m.width= dx;  
		can_m.height= dy;  
			/* the bottom 2 lines of code position the rectangle (left,top) exactly where the user clicked
			can_m.style.left= e.clientX + 'px';  // top and left properties require a unit of measure, not just a number
			can_m.style.top= e.clientY + 'px';  */ 
  
		// if ( can_m.style.display != 'block' )  {  // ? performance
		can_m.style.display= 'block';	
		//}	
		
		can_m.setAttribute('data-cell', pos.join(';'));   // 'data-cell' = 'row;column'  
				
		let ctx= can_m.getContext('2d');
		
			//line bellow copy can_s rectangle part like image..
		ctx.drawImage(can_s, col*dx, row * dy, dx, dy,  0, 0, dx, dy);  //void ctx.drawImage(image, sx, sy, sWidth, sHeight, dx, dy, dWidth, dHeight);

	}	
}


// return row and cell index from mouse coord
// return array [index 0 = row, index 1 = col]. row and col are started from 0 
function getCellFromXY(offsetX, offsetY)
{
  let can= document.getElementById('can_s'), raster= stim[item].length, result= [];
	let dx =  can.width / raster, dy = can.height / raster;  
  let row = Math.floor ( offsetY / dy );  
  let col = Math.floor ( offsetX / dx );  
  result.push(row);
	result.push(col);
  return result;	
}


// ondragstart event for html can_m object 
// html => <canvas id="can_m" draggable="true" ondragstart="dragStart(event)"></canvas> 
function dragStart(e)  {    
	if ( is_task )  {
		e.dataTransfer.setData('text', document.getElementById('can_m').getAttribute('data-cell') );	//  xxxxx  'data-cell' = cell row and column position   xxxxx 
	}
}


// ondragover() event for html canvas element/object
// the ondragover event specifies wher the dragged data can be dropped
// html:  <canvas id="can_s" onmousemove="mouseMove(event)" ondragover="allowDrop(event)" ondrop="drop(event)"></canvas> /
function allowDrop(e)  {
  if ( is_task )  {
		e.preventDefault();
	}
}


// ondrop() event for html canvas element/object
// html: <canvas id="can_s" onmousemove="mouseMove(event)" ondragover="allowDrop(event)" ondrop="drop(event)"></canvas> 
function drop(e)  
{  
  if ( is_task )
	{
		e.preventDefault();   
		let s_data= e.dataTransfer.getData('text');  //  data from start or SOURCE rect => source cell position 	
		s_data = s_data.split(';');
		let s_row= s_data[0], s_col= s_data[1];		
		let d_data= getCellFromXY(e.offsetX, e.offsetY);
		let d_row= d_data[0], d_col= d_data[1];
		 // update (swap) user answer => swap rectangles/cells positions
		let tmp= stim[item][s_row][s_col];
		stim[item][s_row][s_col]= stim[item][d_row][d_col];
		stim[item][d_row][d_col]= tmp;
		
		drawStim();  // repaint stimulus

		if ( auto_f ) 
		{
			let done; 
			if ( stype == '5' ) {
        done = ( isLatin(item) > 0 && boxUnique(item) ) ? true : false;  
      }
      else  {
        done = isLatin(item) > 0 ? true: false;  
      }	
      if (done)  {
        setTimeout(autoNext, 200);  // this time period is not precize
			}				
		}
	}
}


// When the item is successfully completed, feedback is received (Alert dialog box) and the next item is automatically displayed.
function autoNext() 
{
	let s = item < item_cnt ? 'Item is done. Press OK or ENTER for next item.' : 'Task is done.';
	alert(s);
	nextItem();
}


/* return number of points for parameter item. item must be in range [0..item_cnt-1]	
   Each element of the matrix in each row and each column must be different (unique). 
	 return 0 or matrix rows x rows (number of matrix rows x number of matrix column). row count = column count. this is square matrix */
function isLatin(item)
{
	let size = stim[item].length;

	// check matrix by rows. Each element of the matrix in each row must be different (unique). 
  for (let i= 0; i< size; i++)  
	{ 
    let len= stim[item][i].length, v, sel= {};  
	  for (let j= 0; j < len; j++)  {
			v= stim[item][i][j]; 	
      if ( sel[v] == 1 )  {
				return 0;
      }				
			sel[v]= 1;  	
		}   		
	}	

  // check matrix by columns. Each element of the matrix in each column must be different (unique).  
  for (let i= 0; i< size; i++)  
	{ 
    let len= stim[item][i].length, v, sel= {};  
	
		for (let j= 0; j < size; j++)  {						
			v= stim[item][j][i]; 	
      if ( sel[v] == 1 )  {
				return 0;
      }				
			sel[v]= 1;  	
		}   		
	}	
	return ( size * size );
}	


/* only for stype= '5' => Numbers-sudoku.
   check if each of 9 boxes have unique (different) value elements 
	 parameter item must be in range [0..item_cnt-1]
	 return true or false    */
function boxUnique(item)  
{
/*	let box = [ [0,0], [0,3], [0,6], 
								[3,0], [3,3], [3,6], 
								[6,0], [6,3], [6,6] ];  // boxes started at this [row,col]  */
	
	for (let row= 0; row <= 6; row+=3)  
	{
		for (let col= 0; col <= 6; col+=3)  
		{
			let  v, sel= {};
			for (let i= row; i<= row+2; i++)  
			{
				for (let j= col; j<= col+2; j++)  
				{
					v = stim[item][i][j];
					if ( sel[v]  ==  1 )  {
						return false;
					}
          sel[v]= 1;					
				}
			}
		}
	}
	return true;	
}


// end of task (research)
function theEnd()  
{    
	is_task= false;

	document.getElementById('task_section').style.display= 'none';	
  document.getElementById('end').style.display= 'block';   
  document.getElementById('can_s').style.display= 'none';  
  document.getElementById('can_m').style.display= 'none';	
	document.getElementById('result_wrapper').style.display= 'block';
  
	let header= [ 'Item',  'Points', 'Maximum points' ];
	let size= stim.length,  score = 0, max_score = 0, arr = [], rows = [];		
	for (let i = 0; i < size; i++)  {				// for (let i = 0; i < item_cnt; i++)			
		let p = isLatin(i);   // number of points in this i-item
		if ( stype == '5' && p > 0 && boxUnique(i) == false ) {    // check sudoku
      p= 0;  
		}			
		arr.push(p);
		score+= p;  // sum of points in whole task
		let len= stim[i].length;
		max_score+= len * len;
		rows[i]= [ parseInt(i+1,10), p, len * len ]; 
	}  
	rows.push (  [ 'Sum', score,  max_score  ] );   // totals 	
  createTableArray(document.getElementById('table1'), rows, header); 
  	
	// table2 => Descriptive statistics	
	const D = 3;  // number of decimals
	rows = [];  // reset table rows (for table2)
	let n = arr.length;  
	let m =   n > 0  ?  mean(arr).toFixed(D) : '';      
	let sd =  n > 1  ?  stDev(arr, m).toFixed(D) : '';   	 	
  rows.push( [ m, sd, n ] );  // last row
	
	header= ['Average', 'Standard deviation', 'N']; // header		
	createTableArray(document.getElementById('table2'), rows, header); 

  // update 'score' label => above "Repeat" button
  let percent = (100 * score / max_score).toFixed(D);
	document.getElementById('score').textContent= 'Score= ' + score + '. Maximum possible score= ' + max_score + '. Percent= ' + percent + '%';
	
	document.getElementById('l_ts').textContent='Last completed research: ' + new Date();
  document.getElementById('l_res').style.display= 'block';
	
	try 
  {
    if (store) { 
			setStorage(score, max_score, percent, m, sd);   			
    }
  }        
  catch (e)  { 
    document.getElementById('table_grp').style.display= 'block';
		document.getElementById('table_grp').textContent= "Your browser probably doesn't have support for the Web Storage Object or access to the Storage Object is disabled. Check your browser settings. In some browsers (IE11, Edge) Storage Object is not enabled when accessed in a local file (offline)."; 
  }  
}


// return arithmetic mean (mean, average) of input array (arr)
// no validation => input array must be in valid number format	
function mean(arr)  
{	
	let len= arr.length, sum= 0;
	for (let i=0; i<len; i++) {
		sum+= arr[i];
	}
	return (sum/arr.length);
}	


// return Sample Standard Deviation (with Bessel's correction) => N-1
// no validation => input array must be in valid number format
// arr = array argument;  mean = average for input arr 
function stDev(arr, mean) 
{
	let len= arr.length, sum= 0, res;
	for (let i=0; i<len; i++) {
		sum+= Math.pow(arr[i]-mean, 2);	 // Math.pow(base, exponent)
	}
	res= Math.pow(sum/(len-1), 0.5);   // denominator = N-1
	return res;
}


// Formats the date in the desired format. If you do not want this string format do not call this function. use: new Date()
function setDate()  
{
	let today = new Date();
	let dd = today.getDate();
	let mm = today.getMonth()+1; 
	let yyyy = today.getFullYear();
	if (dd < 10) { dd='0'+dd; } 
	if (mm < 10) { mm='0'+mm; } 
	today = dd+'.' + mm + '.' +yyyy;
	return today;
}


/* Store the results of the last completed task in the user's browser
   set user data into window.localStorage. Parameters:
	 score = sum (number) of points in task
   max_score = maximum possible points in task
   percent = 100 * points/ max_points	 
   mean = average (mean) for user points 
   sd = standard deviation for user points	 */  
function setStorage(score, max_score, percent, mean, sd)  
{    
	let oldItems = JSON.parse(localStorage.getItem('lst-v1.0')) || []; 
  let id= oldItems.length + 1;     	
	// set new "record"
  let newItem = {
      'id' : id,
      'date' : setDate(),   //new Date(),  		   
			'score' : score,
			'max-score' : max_score,
			'percent' : percent,
			'mean'  : mean, 	
			'sd' : sd,			
		  'stimulus' : stype,
			'row-min' : row_min,
      'row-max' : row_max,			
			'iterations' : iter,
			'items' : item_cnt,		
			'timer' : timer_enabled ? timer : '',	
			'is-back' : is_back ? 1 : 0,
			'auto-f'  : auto_f ? 1 : 0,
			'shuffle' : is_shuffle ? 1 : 0,
			'demo' : demo ? 1 : 0 
  };

  oldItems.push(newItem);

  localStorage.setItem('lst-v1.0', JSON.stringify(oldItems));
  
	document.getElementById('table_grp').style.display= 'block';
  document.getElementById('store_id').style.display= 'block';
  document.getElementById('store_id').textContent= 'The results are stored in your browser. ID= ' + id ;
  
	createTableJSON(document.getElementById('table_grp'), oldItems); 
}


// convert JSON data dynamically to HTML Table using JavaScript - Dynamically Create table with rows in JavaScript
// downloaded and adapted from source: https://www.encodedna.com/javascript/populate-json-data-to-html-table-using-javascript.htm
function createTableJSON(elem, my_json) 
{  
  // EXTRACT VALUE FOR HTML HEADER. 
  let col = [];
  for (let i = 0; i < my_json.length; i++) {
    for (let key in my_json[i]) {
        if (col.indexOf(key) === -1) {
          col.push(key);
        }
    }
  } 
  let table = document.createElement('table');
  // CREATE HTML TABLE HEADER ROW USING THE EXTRACTED HEADERS ABOVE.
  let tr = table.insertRow(-1);                   // TABLE ROW.
  for (let i = 0; i < col.length; i++) {
      let th = document.createElement('th');      // TABLE HEADER.
      th.innerHTML = col[i];
      tr.appendChild(th);
  }
  // ADD JSON DATA TO THE TABLE AS ROWS.
  for (let i = 0; i < my_json.length; i++) 
  {
    tr = table.insertRow(-1);
    for (let j = 0; j < col.length; j++) {
      let tabCell = tr.insertCell(-1);
      tabCell.innerHTML = my_json[i][col[j]];
    }
  }
  // FINALLY ADD THE NEWLY CREATED TABLE WITH JSON DATA TO A CONTAINER.
  elem.innerHTML= '';
  elem.appendChild(table);
}


/* convert array data dynamically to HTML table - dynamically create html table
   elem == DOM html element (node)
   data => 1-dim or 2-dim array (matrix)
   col => string column descriptions. Set an empty string if you want to skip a column inside a 2-dim array	 */ 
function createTableArray(elem, data, col) 
{  
	let table= document.createElement('table');  // html table
  let tr= table.insertRow(-1), th, cell, tmp;  // table row, table header, table cell 	
  // create html table header row using the extracted headers above  
  for (let i in col) {
		if ( col[i] == '' ) { continue; } // set an empty string if you want to skip a column inside a 2-dim array
		th= document.createElement('th');      
		th.innerHTML = col[i];
		tr.appendChild(th);
  }
  if ( is2dArray(data) ) 
	{ // add 2-dim array data to the table as rows		
		tmp= data.length;
		for (let i = 0; i < tmp; i++) 
		{
			tr= table.insertRow(-1);
			for (let j in col) {
				if (col[j] == '') { continue; }  // set an empty string if you want to skip a column inside a 2-dim array
				cell= tr.insertCell(-1);
				cell.innerHTML = data[i][j];
			}
		}
  }
	else 
	{	// add 1-dim array data to the table as only one row
		tr= table.insertRow(-1);
		tmp=data.length;
		for (let i = 0; i < tmp; i++)	{
			cell= tr.insertCell(-1);
			cell.innerHTML = data[i];
		}
	}	
  elem.innerHTML= '';
  elem.appendChild(table); // finally add the newly created table with string data to a html container
}


// test if array is 2-dim (return true) or not (return false)
function is2dArray(arr)
{
	if(arr[0] === undefined){
		return false;
	}
	else	{
		return (arr[0].constructor === Array);
	}
}


// clear web browser storage data for this application
function clrStorage()  
{  
  let cnt= localStorage.getItem('lst-v1.0')  ?  JSON.parse(localStorage.getItem('lst-v1.0')).length : 0;                                                        
  if (cnt === 0)  {
    //alert('You have no data to delete');
    return;
  }  
  if ( !window.confirm('You have a total of ' + cnt  + ' results stored in your web browser. Continue?') ) { 
   return;
 }                                             
 if (window.confirm('Delete total saved results:  '  + cnt + '. This action cannot be undone. Final confirmation?') ) 
 { 
   window.localStorage.removeItem('lst-v1.0'); 
   document.getElementById('store_cnt').textContent= 'All data has been deleted. No data.';
	 document.getElementById('store_oper').style.display= 'none';
 }  
}


// copy data from user web browser to the Clipboard 
function copyStorage()  
{  
  let data= localStorage.getItem('lst-v1.0');
	if (!data)  {
    //alert('You have no data to Copy');
    return;
  }  
	data = JSON.parse(data); 
	let arr= jsonToArray(data);	
	copy2DToClipboard(arr);
}	


// convert json object into 2D array
// data = JSON object
// return 2D array
function jsonToArray(data)  
{
   // extract column (fields) names  
  let total= data.length, col = [], row, res=[];  // res= result (2d array) 
  for (let i = 0; i < total; i++) {
    for (let key in data[i]) {
			if (col.indexOf(key) === -1)  {
				col.push(key);  // col= columns (fields) names
			}
    }
  } 
	res.push(col);  // add field names as 1 row
  let c = col.length;    // column count  
  // add JSON data to the 2d array as rows.
  for (let i = 0; i < total; i++) 
  {
		row= [];
    for (let j = 0; j < c; j++)  {     
      row.push(data[i][col[j]]);
    }
		res.push(row);
  }
	return res;	
}


// https://stackoverflow.com/questions/60698115/how-to-copy-a-javascript-2d-array-to-clipboard-to-paste-it-in-excel
// copy 2D array to Clipboard
function copy2DToClipboard(arr) {
  let csv = '', row, cell;
  for (row = 0; row < arr.length; row++) {
    for (cell = 0; cell < arr[row].length; cell++) {
      csv += (arr[row][cell]+'').replace(/[\n\t]+/g, ' ');
      if (cell+1 < arr[row].length) csv += '\t';
    }
    if (row+1 < arr.length) { 
			csv += '\n'; 
		}
  }
  copyTextToClipboard(csv);
}


// copy 2D array (or text) to Clipboard
function copyTextToClipboard(txt) 
{	
	if (!navigator.clipboard) {
    fallbackCopy(txt);
    return;
  }
  navigator.clipboard.writeText(txt).then(function() {
   //alert('Async: Copying to Clipboard was successful!');
  }, function(err) {
    alert('Oops, could not copy data. Error: ' + err);
  });
}


// copied from https://stackoverflow.com/questions/400212/how-do-i-copy-to-the-clipboard-in-javascript
// fallback copy text to clipboard
function fallbackCopy(txt) 
{
  let ta = document.createElement('textarea');
  ta.value = txt; 
  // Avoid scrolling to bottom
  ta.style.top = '0';
  ta.style.left = '0';
  ta.style.position = 'fixed';
  document.body.appendChild(ta);
  ta.focus();
  ta.select();
  try {
    let successful = document.execCommand('copy');
    let msg = successful ? 'successful' : 'unsuccessful';
  } catch (err) {
    alert('Oops, unable to copy. Error: ' + err);
  }
  document.body.removeChild(ta);
}


// https://stackoverflow.com/questions/2004555/how-can-i-make-javascript-make-produce-new-page
// displays user saved results (Window.localStorage application data) in new empty web page (tab) 
function newWin()
{ 
	let data = JSON.parse(localStorage.getItem('lst-v1.0'))
	createTableJSON(document.getElementById('table_grp'), data);
	let stl= '<style>table{margin:1em auto;border:thick solid #000;border-collapse:collapse}td,th{border:medium solid #000;text-align:center;padding:.5em}th{font-weight:bold}</style>';
	let tbl= document.getElementById('table_grp').innerHTML;	
	let win = window.open('');
  win.document.write('<html><head><title>Saved data</title>' + stl + '</head><body>' + tbl + '</body></html>');
  win.focus();
}


// https://stackoverflow.com/questions/2004555/how-can-i-make-javascript-make-produce-new-page
// displays Latest research results in new empty web page (tab) 
// parameter: id => DOM/html object (div or table)
function newWinLast(id)
{ 	
	let ind = document.getElementById('last_data_sel').selectedIndex;
	ind++;
	let e = document.getElementById('table' + ind);	
	let stl= '<style>table{margin:1em auto;border:thick solid #000;border-collapse:collapse}td,th{border:medium solid #000;text-align:center;padding:.5em}th{font-weight:bold}</style>';
	let tbl= e.innerHTML; 	
	let win = window.open('');
	win.document.write('<html><head><title>Latest research data</title>' + stl + '</head><body>' + tbl + '</body></html>');
	win.focus();
}


// copy only latest task results (data) to the Clipboard 
function copyLast()  
{ 
  let ind = document.getElementById('last_data_sel').selectedIndex, tbl;
  switch (ind) {
    case 0:
			tbl = document.getElementById('table1').getElementsByTagName('table')[0];  // note: id='table1' is html div element
			break;		
    case 1:
			tbl = document.getElementById('table2').getElementsByTagName('table')[0];  
			break;
	}		
  if (tbl)  {
		let data= table2DArray(tbl);
		if (data.length > 0) {
			copy2DToClipboard(data);
		}
	}
}	


// https://stackoverflow.com/questions/3065342/how-do-i-iterate-through-table-rows-and-cells-in-javascript
// input argument tbl == html/DOM  table  object
// return 2D array
function table2DArray(tbl) 
{
	let res = [];	 
	for (let i = 0, row; row = tbl.rows[i]; i++) 
	{   //iterate through rows. rows would be accessed using the "row" variable assigned in the for loop
	  res[i]= [];		
		for (let j = 0, col; col = row.cells[j]; j++) {  //iterate through columns. columns would be accessed using the "col" variable assigned in the for loop			 
		  res[i][j]= col.textContent;
		}  
	}
	return res;				
}
